open Syntax
open Value

(* 実際の計算をする関数 *)
(* Interpreter with linear instructions : eval10 *)

(* cons : c -> s -> t -> t *)
let rec cons c s t = match t with
    TNil -> Trail (c, s)
  | Trail (c', s') -> Trail (CAppend (c, c'), SAppend (s, s'))

(* apnd : t -> t -> t *)
let apnd t0 t1 = match t0 with
    TNil -> t1
  | Trail (c, s) -> cons c s t1

(* at : c -> c -> c *)
let rec at c0 c1 = match c0 with
    CCons (i, CNil) -> CCons (i, c1)
  | CCons (i, c) -> CCons (i, at c c1)
  | _ -> failwith "cont error"

(* run_c10 : c -> s -> t -> m -> v *)
let rec run_c10 c s t m = match (c, s) with
    (CNil, SCons (v, SNil)) ->
    begin match t with
        TNil ->
        begin match m with
            [] -> v
          | (c, s, t) :: m -> run_c10 c (SCons (v, s)) t m
        end
      | Trail (c, s) -> run_c10 c (SCons (v, s)) TNil m
    end
  | (CCons (INum (v), c), SCons (VEnv (vs), s)) -> run_c10 c (SCons (v, s)) t m
  | (CCons (IAccess (n), c), SCons (VEnv (vs), s)) ->
    run_c10 c (SCons (List.nth vs n, s)) t m
  | (CCons (IPush_closure (i), c), SCons (VEnv (vs), s)) ->
    run_c10 c (SCons (VFun (i, vs), s)) t m
  | (CCons (IReturn, _), SCons (v, SCons (VK (c), s))) ->
    run_c10 c (SCons (v, s)) t m
  | (CCons (IPush_env, c), SCons (VEnv (vs), s)) ->
    run_c10 c (SCons (VEnv (vs), SCons (VEnv (vs), s))) t m
  | (CCons (IPop_env, c), SCons (v, SCons (VEnv (vs), s))) ->
    run_c10 c (SCons (VEnv (vs), SCons (v, s))) t m
  | (CCons (IOperations (op), c), SCons (v1, SCons (v0, s))) ->
    begin match (v0, v1) with
        (VNum (n0), VNum (n1)) ->
        begin match op with
            Plus -> run_c10 c (SCons (VNum (n0 + n1), s)) t m
          | Minus -> run_c10 c (SCons (VNum (n0 - n1), s)) t m
          | Times -> run_c10 c (SCons (VNum (n0 * n1), s)) t m
          | Divide ->
            if n1 = 0 then failwith "Division by zero"
            else run_c10 c (SCons (VNum (n0 / n1), s)) t m
        end
      | _ -> failwith "op error"
    end
  | (CCons (ICall, c), SCons (v1, SCons (v0, s))) ->
    begin match v0 with
        VFun (c', vs) ->
        run_c10 c' (SCons (VEnv (v1 :: vs), SCons (VK (c), s))) t m
      | VContS (c', s', t') ->
        run_c10 c' (SCons (v1, s')) t' ((c, s, t) :: m)
      | VContC (c', s', t') ->
        run_c10 c' (SCons (v1, s')) (apnd t' (cons c s t)) m
      | _ -> failwith "not app"
    end
  | (CCons (IShift (c'), c), SCons (VEnv (vs), s)) ->
    run_c10 c' (SCons (VEnv (VContS (c, s, t) :: vs), SNil)) TNil m
  | (CCons (IControl (c'), c), SCons (VEnv (vs), s)) ->
    run_c10 c' (SCons (VEnv (VContC (c, s, t) :: vs), SNil)) TNil m
  | (CCons (IShift0 (c'), c), SCons (VEnv (vs), s)) ->
    begin match m with
        (c0, s0, t0) :: m0 ->
        run_c10 (at c' c0) (SCons (VEnv (VContS (c, s, t) :: vs), s0)) t0 m0
      | _ -> failwith "meta cont error"
    end
  | (CCons (IControl0 (c'), c), SCons (VEnv (vs), s)) ->
    begin match m with
        (c0, s0, t0) :: m0 ->
        run_c10 (at c' c0) (SCons (VEnv (VContC (c, s, t) :: vs), s0)) t0 m0
      | _ -> failwith "meta cont error"
    end
  | (CCons (IReset (c'), c), SCons (VEnv (vs), s)) ->
    run_c10 c' (SCons (VEnv (vs), SNil)) TNil ((c, s, t) :: m)
  | (CAppend (c, c'), SCons (v, SAppend (s, s'))) ->
    run_c10 c (SCons (v, s)) (cons c' s' t) m
  | _ -> failwith "cont or stack error"

(* f10 : e -> string list -> c *)
let rec f10 e xs = match e with
    Num (n) -> CCons (INum (VNum (n)), CNil) 
  | Var (x) -> CCons (IAccess (Env.offset x xs), CNil)
  | Op (e0, op, e1) ->
    at
      (CCons (IPush_env, CNil))
      (at
         (f10 e0 xs)
         (at
            (CCons (IPop_env, CNil))
            (at
               (f10 e1 xs)
               (CCons (IOperations (op), CNil)))))
  | Fun (x, e) ->
    CCons (IPush_closure (at (f10 e (x :: xs)) (CCons (IReturn, CNil))), CNil)
  | App (e0, e1) ->
    at
      (CCons (IPush_env, CNil))
      (at
         (f10 e0 xs)
         (at
            (CCons (IPop_env, CNil))
            (at
               (f10 e1 xs)
               (CCons (ICall, CNil)))))
  | Shift (x, e) -> CCons (IShift (f10 e (x :: xs)), CNil)
  | Control (x, e) -> CCons (IControl (f10 e (x :: xs)), CNil)
  | Shift0 (x, e) -> CCons (IShift0 (f10 e (x :: xs)), CNil)
  | Control0 (x, e) -> CCons (IControl0 (f10 e (x :: xs)), CNil)
  | Reset (e) -> CCons (IReset (f10 e xs), CNil)
           
(* f : e -> v *)
let f expr = run_c10 (f10 expr []) (SCons (VEnv ([]), SNil)) TNil []
