open Syntax
open Value

(* 実際の計算をする関数 *)
(* Defunctionalized interpreter : eval2 *)

(* 初期継続 *)
let idc = C0

(* cons : c -> t -> t *)
let cons c t = match t with
    TNil -> Trail (c)
  | Trail (c') -> Trail (CAppend (c, c'))

(* apnd : t -> t -> t *)
let apnd t0 t1 = match t0 with
    TNil -> t1
  | Trail (c) -> cons c t1

(* run_c2 : c -> v -> t -> m -> v *)
let rec run_c2 c v t m = match c with
    C0 ->
    begin match t with
        TNil ->
        begin match m with
            [] -> v
          | (c, t) :: m -> run_c2 c v t m
        end
      | Trail (c) -> run_c2 c v TNil m
    end
  | CApp0 (e1, xs, vs, c) -> f2 e1 xs vs (CApp1 (v, c)) t m
  | CApp1 (v0, c) ->
    begin match v0 with
        VFun (e, x, xs, vs) -> f2 e (x :: xs) (v :: vs) c t m
      | VContS (c', t') -> run_c2 c' v t' ((c, t) :: m)
      | VContC (c', t') -> run_c2 c' v (apnd t' (cons c t)) m
      | _ -> failwith "not app"
    end
  | COp0 (e1, xs, vs, op, c) -> f2 e1 xs vs (COp1 (v, op, c)) t m 
  | COp1 (v0, op, c) ->
    begin match (v0, v) with
        (VNum (n0), VNum (n1)) ->
        begin match op with
            Plus -> run_c2 c (VNum (n0 + n1)) t m
          | Minus -> run_c2 c (VNum (n0 - n1)) t m
          | Times -> run_c2 c (VNum (n0 * n1)) t m
          | Divide ->
            if n1 = 0 then failwith "Division by zero"
            else run_c2 c (VNum (n0 / n1)) t m
        end
      | _ -> failwith "op error"
    end
  | CAppend (c, c') -> run_c2 c v (cons c' t) m

(* f2 : e -> string list -> v list -> c -> t -> m -> v *)
and f2 e xs vs c t m = match e with
    Num (n) -> run_c2 c (VNum (n)) t m 
  | Var (x) -> run_c2 c (List.nth vs (Env.offset x xs)) t m
  | Op (e0, op, e1) -> f2 e0 xs vs (COp0 (e1, xs, vs, op, c)) t m
  | Fun (x, e) -> run_c2 c (VFun (e, x, xs, vs)) t m
  | App (e0, e1) -> f2 e0 xs vs (CApp0 (e1, xs, vs, c)) t m
  | Shift (x, e) -> f2 e (x :: xs) (VContS (c, t) :: vs) idc TNil m
  | Control (x, e) -> f2 e (x :: xs) (VContC (c, t) :: vs) idc TNil m
  | Shift0 (x, e) ->
    begin match m with
        (c0, t0) :: m0 -> f2 e (x :: xs) (VContS (c, t) :: vs) c0 t0 m0
      | _ -> failwith "meta cont error"
    end
  | Control0 (x, e) ->
    begin match m with
        (c0, t0) :: m0 -> f2 e (x :: xs) (VContC (c, t) :: vs) c0 t0 m0
      | _ -> failwith "meta cont error"
    end
  | Reset (e) -> f2 e xs vs idc TNil ((c, t) :: m)

(* f : e -> v *)
let f expr = f2 expr [] [] idc TNil []
