open Syntax
open Value

(* 実際の計算をする関数 *)
(* Delinearized interpreter : eval5 *)

(* 初期継続 *)
let idc = C0

(* cons : c -> s -> t -> t *)
let cons c s t = match t with
    TNil -> Trail (c, s)
  | Trail (c', s') -> Trail (CAppend (c, c'), SAppend (s, s'))

(* apnd : t -> t -> t *)
let apnd t0 t1 = match t0 with
    TNil -> t1
  | Trail (c, s) -> cons c s t1

(* run_c5 : c -> v -> s -> t -> m -> v *)
let rec run_c5 c v s t m = match (c, s) with
    (C0, SNil) ->
    begin match t with
        TNil ->
        begin match m with
            [] -> v
          | (c, s, t) :: m -> run_c5 c v s t m
        end
      | Trail (c, s) -> run_c5 c v s TNil m
    end
  | (CApp0 (e1, xs, c), SCons (VEnv (vs), s)) ->
    f5 e1 xs vs (SCons (v, s)) (CApp1 (c)) t m
  | (CApp1 (c), SCons (v0, s)) ->
    begin match v0 with
        VFun (e, x, xs, vs) -> f5 e (x :: xs) (v :: vs) s c t m
      | VContS (c', s', t') -> run_c5 c' v s' t' ((c, s, t) :: m)
      | VContC (c', s', t') -> run_c5 c' v s' (apnd t' (cons c s t)) m
      | _ -> failwith "not app"
    end
  | (COp0 (e1, xs, op, c), SCons (VEnv (vs), s)) ->
    f5 e1 xs vs (SCons (v, s)) (COp1 (op, c)) t m 
  | (COp1 (op, c), SCons (v0, s)) ->
    begin match (v0, v) with
        (VNum (n0), VNum (n1)) ->
        begin match op with
            Plus -> run_c5 c (VNum (n0 + n1)) s t m
          | Minus -> run_c5 c (VNum (n0 - n1)) s t m
          | Times -> run_c5 c (VNum (n0 * n1)) s t m
          | Divide ->
            if n1 = 0 then failwith "Division by zero"
            else run_c5 c (VNum (n0 / n1)) s t m
        end
      | _ -> failwith "op error"
    end
  | (CAppend (c, c'), SAppend (s, s')) -> run_c5 c v s (cons c' s' t) m
  | _ -> failwith "stack and cont lengths are different"

(* f5 : e -> string list -> v list -> s -> c -> t -> m -> v *)
and f5 e xs vs s c t m = match e with
    Num (n) -> run_c5 c (VNum (n)) s t m 
  | Var (x) -> run_c5 c (List.nth vs (Env.offset x xs)) s t m
  | Op (e0, op, e1) ->
    f5 e0 xs vs (SCons (VEnv (vs), s)) (COp0 (e1, xs, op, c)) t m
  | Fun (x, e) -> run_c5 c (VFun (e, x, xs, vs)) s t m
  | App (e0, e1) ->
    f5 e0 xs vs (SCons (VEnv (vs), s)) (CApp0 (e1, xs, c)) t m
  | Shift (x, e) -> f5 e (x :: xs) (VContS (c, s, t) :: vs) SNil idc TNil m
  | Control (x, e) -> f5 e (x :: xs) (VContC (c, s, t) :: vs) SNil idc TNil m
  | Shift0 (x, e) ->
    begin match m with
        (c0, s0, t0) :: m0 ->
        f5 e (x :: xs) (VContS (c, s, t) :: vs) s0 c0 t0 m0
      | _ -> failwith "meta cont error"
    end
  | Control0 (x, e) ->
    begin match m with
        (c0, s0, t0) :: m0 ->
        f5 e (x :: xs) (VContC (c, s, t) :: vs) s0 c0 t0 m0
      | _ -> failwith "meta cont error"
    end
  | Reset (e) -> f5 e xs vs SNil idc TNil ((c, s, t) :: m)

(* f : e -> v *)
let f expr = f5 expr [] [] SNil idc TNil []
