open Syntax
open Value

(* 実際の計算をする関数 *)
(* Interpreter with combined arguments : eval7 *)

(* 初期継続 *)
let idc s t m = match s with
    SCons (v, SNil) ->
    begin match t with
        TNil ->
        begin match m with
            MNil -> v
          | MCons ((c, s, t) ,m) -> c (SCons (v, s)) t m
        end
      | Trail (c, s) -> c (SCons (v, s)) TNil m
    end
  | _ -> failwith "stack error"

(* cons : c -> s -> t -> t *)
let rec cons c s t = match t with
    TNil -> Trail (c, s)
  | Trail (c', s') ->
    Trail (
      (fun s t m ->
         begin match s with
             (SCons (v, SAppend (s, s'))) -> c (SCons (v, s)) (cons c' s' t) m
           | _ -> failwith "stack error"
         end),
      SAppend (s, s'))

(* apnd : t -> t -> t *)
let apnd t0 t1 = match t0 with
    TNil -> t1
  | Trail (c, s) -> cons c s t1

(* f7 : e -> string list -> s -> c -> t -> m -> v *)
let rec f7 e xs s c t m = match s with
    SCons (VEnv (vs), s) ->
    begin match e with
        Num (n) -> c (SCons (VNum (n), s)) t m 
      | Var (x) -> c (SCons (List.nth vs (Env.offset x xs), s)) t m
      | Op (e0, op, e1) ->
        f7 e0 xs (SCons (VEnv (vs), SCons (VEnv (vs), s))) (fun s0 t0 m0 ->
            begin match s0 with
                SCons (v0, SCons (VEnv (vs), s0)) ->
                f7 e1 xs (SCons (VEnv (vs), SCons (v0, s0))) (fun s1 t1 m1 ->
                    begin match s1 with
                        SCons (v1, SCons (v0, s1)) ->
                        begin match (v0, v1) with
                            (VNum (n0), VNum (n1)) ->
                            begin match op with
                                Plus -> c (SCons (VNum (n0 + n1), s1)) t1 m1
                              | Minus -> c (SCons (VNum (n0 - n1), s1)) t1 m1
                              | Times -> c (SCons (VNum (n0 * n1), s1)) t1 m1
                              | Divide ->
                                if n1 = 0 then failwith "Division by zero"
                                else c (SCons (VNum (n0 / n1), s1)) t1 m1
                            end
                          | _ -> failwith "op error"
                        end
                      | _ -> failwith "stack error"
                    end) t0 m0
              | _ -> failwith "stack error"
            end) t m
      | Fun (x, e) ->
        let vfun = VFun (fun s' c' t' m' ->
            begin match s' with
                SCons (v, s') ->
                f7 e (x :: xs) (SCons (VEnv (v :: vs), s')) c' t' m'
              | _ -> failwith "stack error"
            end) in
        c (SCons (vfun, s)) t m
      | App (e0, e1) ->
        f7 e0 xs (SCons (VEnv (vs), SCons (VEnv (vs), s))) (fun s0 t0 m0 ->
            begin match s0 with
                SCons (v0, SCons (VEnv (vs), s0)) ->
                f7 e1 xs (SCons (VEnv (vs), SCons (v0, s0))) (fun s1 t1 m1 ->
                    begin match s1 with
                        SCons (v1, SCons (v0, s1)) ->
                        begin match v0 with
                            VFun (f) -> f (SCons (v1, s1)) c t1 m1
                          | VContS (c', s', t') ->
                            c' (SCons (v1, s')) t' (MCons ((c, s1, t1), m1))
                          | VContC (c', s', t') ->
                            c' (SCons (v1, s')) (apnd t' (cons c s1 t1)) m1
                          | _ -> failwith "not app"
                        end
                      | _ -> failwith "stack error"
                    end) t0 m0
              | _ -> failwith "stack error"
            end) t m
      | Shift (x, e) ->
        f7 e (x :: xs) (SCons (VEnv (VContS (c, s, t) :: vs), SNil)) idc TNil m
      | Control (x, e) ->
        f7 e (x :: xs) (SCons (VEnv (VContC (c, s, t) :: vs), SNil)) idc TNil m
      | Shift0 (x, e) ->
        begin match m with
            MCons ((c0, s0, t0), m0) ->
            f7 e (x :: xs) (SCons (VEnv (VContS (c, s, t) :: vs), s0)) c0 t0 m0
          | _ -> failwith "meta cont error"
        end
      | Control0 (x, e) ->
        begin match m with
            MCons ((c0, s0, t0), m0) ->
            f7 e (x :: xs) (SCons (VEnv (VContC (c, s, t) :: vs), s0)) c0 t0 m0
          | _ -> failwith "meta cont error"
        end
      | Reset (e) ->
        f7 e xs (SCons (VEnv (vs), SNil)) idc TNil (MCons ((c, s, t), m))
    end
  | _ -> failwith "stack error"
           
(* f : e -> v *)
let f expr = f7 expr [] (SCons (VEnv ([]), SNil)) idc TNil MNil
