open Syntax

(* Interpreter with defunctionalized continued, instructions : eval9 *)

(* Value : プログラムの実行結果を表す型 *)
type v = VNum of int
       | VFun of i * v list
       | VContS of c * s * t
       | VContC of c * s * t
       | VEnv of v list
       | VK of c

and i = INum of v | IAccess of int
      | IPush_closure of i | IReturn
      | IPush_env | IPop_env
      | IOperations of op | ICall
      | IShift of i | IControl of i
      | IShift0 of i | IControl0 of i
      | IReset of i
      | ISeq of i * i

and c = CNil
      | CCons of i * c
      | CAppend of c * c

and s = SNil
      | SCons of v * s
      | SAppend of s * s

and t = TNil | Trail of c * s

and m = (c * s * t) list

(* プログラムの実行結果を文字列にする関数 *)
(* to_string : v -> string *)
let rec to_string value = match value with
    VNum (n) -> string_of_int n
  | VFun (_) -> "<VFun>"
  | VContS (_) -> "<VContS>"
  | VContC (_) -> "<VContC>"
  | VEnv (_) -> "<VEnv>"
  | VK (_) -> "<VK>"

(* プログラムの実行結果をプリントする関数 *)
(* Value.print : v -> unit *)
let print exp =
  let str = to_string exp in
  print_string str
